# -*- coding: utf-8 -*-
#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from ModelClasses import *
import PlotStyles

import re

from Tigger.Coordinates import angular_dist_pos_angle,DEG

class ModelTag (ModelItem):
  mandatory_attrs = [ "name" ];
  optional_attrs = dict([ (attr,None) for attr in PlotStyles.StyleAttributes ]);

ModelTag.registerClass();

class ModelTagSet (ModelItem):
  def __init__ (self,*tags,**kws):
    ModelItem.__init__(self,**kws);
    self.tags = dict([ (tag.name,tag) for tag in tags ]);

  def add (self,tag):
    """Adds a ModelTag object to the tag set""";
    self.tags[tag.name] = tag;

  def get (self,tagname):
    """Returns ModelTag object associated with tag name, inserting a new one if not found""";
    return self.tags.setdefault(name,ModelTag(name));

  def getAll (self):
    all = self.tags.values();
    all.sort(lambda a,b:cmp(a.name,b.name));
    return all;

  def addNames (self,names):
    """Ensures that ModelTag objects are initialized for all tagnames in names""";
    for name in names:
      self.tags.setdefault(name,ModelTag(name));

  def renderMarkup (self,tag="A",attrname=None):
      """Makes a markup string corresponding to the model item.
      'tags' is the HTML tag to use.
      """;
      # opening tag
      markup = "<%s mdltype=ModelTagList "%tag;
      if attrname is not None:
        markup += "mdlattr=%s "%attrname;
      markup +=">";
      # write mandatory attributes
      for name,tt in self.tags.iteritems():
        markup += self.renderAttrMarkup(name,tt,tag="TR",mandatory=True);
      # closing tag
      markup += "</%s>"%tag;
      return markup;
ModelTagSet.registerClass();

class Source (ModelItem):
  """Source represents a model source.
  Each source has mandatory name (class str), pos (class Position) and flux (class Flux) model attributes.
  There are optional spectrum (class Spectrum) and shape (class Shape) model attributes.

  Standard Python attributes of a Source object are:
    selected: if the source is selected (e.g. in a selection widget)
    typecode: a type code. This is "pnt" if no shape is set (i.e.for a delta-function), otherwise it's the shape's typecode.
  """;
  mandatory_attrs  = [ "name","pos","flux" ];
  optional_attrs     = dict(spectrum=None,shape=None);
  allow_extra_attrs = True;

  def __init__ (self,*args,**kw):
    ModelItem.__init__(self,*args,**kw);
    self.typecode = (self.shape and self.shape.typecode) or "pnt";
    self.selected = False;

  def select (self,sel=True):
    self.selected = sel;

  def brightness (self):
    iapp = getattr(self,'Iapp',None);
    if iapp is not None:
      return iapp;
    else:
      return getattr(self.flux,'I',0.);

  def get_attr (self,attr,default=None):
    return getattr(self,attr,default);

  def getTagNames (self):
    return [ attr for attr,val in self.getExtraAttributes() if attr[0] != "_" ];

  def getTags (self):
    return [ (attr,val) for attr,val in self.getExtraAttributes() if attr[0] != "_" ];

  getTag = get_attr;
  setTag = ModelItem.setAttribute;

  class Grouping (object):
    # show_plot settings
    NoPlot = 0;
    Default = 1;
    Plot = 2;
    def __init__ (self,name,func,style=PlotStyles.DefaultPlotStyle,sources=None):
      self.name = name;
      self.style = style;
      self.func = func;
      self.total = 0;
      if sources:
        self.computeTotal(sources);
    def computeTotal (self,sources):
      self.total = len(filter(self.func,sources));
      return self.total;

Source.registerClass();

class SkyModel (ModelItem):
  optional_attrs   = dict(name=None,plotstyles={},pbexp=None,ra0=None,dec0=None,freq0=None);
  allow_extra_attrs = True;

  def __init__ (self,*sources,**kws):
    ModelItem.__init__(self,**kws);
    # "current" source (grouping "current" below is defined as that one source)
    self._current_source = None;
    self._filename = None;
    # list of loaded images associated with this model
    self._images = [];
    # setup source list
    self.setSources(sources);

  def copy (self):
    return SkyModel(*self.sources,**dict(self.getAttributes()));

  def images (self):
    """Returns list of images associated with this model""";
    return self._images;

  def setFilename (self,filename):
    self._filename = filename;

  def filename (self):
    return self._filename;

  def setCurrentSource (self,src,origin=None):
    """Changes the current source. If it has indeed changed, emits a currentSourceChanged signal. Arguments passed with the signal:
    src: the new current source.
    src0: the previously current source.
    origin: originator of changes.
    """;
    if self._current_source is not src:
      src0 = self._current_source;
      self._current_source = src;
      if self.signalsEnabled():
        self.emit("changeCurrentSource",src,src0,origin);

  def currentSource (self):
    return self._current_source;

  # Bitflags for the 'what' argument of the updated() signal below.
  # These indicate what exactly has been updated:
  UpdateSourceList  =  1;           # source list changed
  UpdateSourceContent = 2;      # source attributes have changed
  UpdateTags = 4;                     # tags have been changed
  UpdateGroupVis = 8;               # visibility of a grouping (group.style.show_list attribute) has changed
  UpdateGroupStyle = 16;          # plot style of a grouping has changed
  UpdateSelectionOnly = 32;       # (in combination with UpdateSourceContent): update only affects currently selected sources
  UpdateAll =  UpdateSourceList +UpdateSourceContent+UpdateTags+UpdateGroupVis+UpdateGroupStyle ;

  def emitUpdate (self,what=UpdateSourceContent,origin=None):
    """emits an updated() signal, indicating that the model has changed. Arguments passed through with the signal:
    what: what is updated. A combination of flags above.
    origin: originator of changes.
    """
    if self.signalsEnabled():
      self.emit("updated",what,origin);

  def emitSelection (self,origin=None):
    """emits an selected() signal, indicating that the selection has changed. Arguments passed through with the signal:
    num: number of selected sources.
    origin: originator of changes.
    """;
    self.selgroup.computeTotal(self.sources);
    if self.signalsEnabled():
      self.emit("selected",self.selgroup.total,origin);

  def emitChangeGroupingVisibility (self,group,origin=None):
    if self.signalsEnabled():
      self.emit("changeGroupingVisibility",group,origin);
      self.emitUpdate(SkyModel.UpdateGroupVis,origin);

  def emitChangeGroupingStyle (self,group,origin=None):
    if self.signalsEnabled():
      self.emit("changeGroupingStyle",group,origin);
      self.emitUpdate(SkyModel.UpdateGroupStyle,origin);

  def findSource (self,name):
    return self._src_by_name[name];

  def setSources (self,sources,origin=None,recompute_r=False):
    # if recompute_r is True, recomputes the 'r' attribute for all sources
    self.sources = list(sources);
    self._src_by_name = dict([(src.name,src) for src in self.sources]);
    if recompute_r:
      self.recomputeRadialDistance();
    self.scanTags();
    self.initGroupings();

  def addSources (self,sources,recompute_r=True):
    # if recompute_r is True, recomputes the 'r' attribute for new sources
    if recompute_r:
      self.recomputeRadialDistance(sources);
    self.setSources(list(self.sources)+list(sources));
    
  def __len__ (self):
    return len(self.sources);
    
  def __getitem__ (self,key):
    if isinstance(key,(int,slice)):
      return self.sources[key];
    elif isinstance(key,str):
      return self.findSource(key);
    else:
      raise TypeError("cannot index SkyModel with key of type %s"%str(type(key)));
    
  def __setitem__ (self,key,value):
    raise TypeError("cannot assign to items of SkyModel, use the setSources() method instead");
  
  def __iter__ (self):
    return iter(self.sources);

  def recomputeRadialDistance (self,sources=None):
    # refreshes the radial distance for a group of sources, or all sources in the model
    if (self.ra0 and self.dec0) is not None:
      for src in (sources or self.sources):
        r,pa = angular_dist_pos_angle(src.pos.ra,src.pos.dec,self.ra0,self.dec0);
        src.setAttribute('r',r);

  def scanTags (self,sources=None):
    """Populates self.tagnames with a list of tags present in sources""";
    sources = sources or self.sources;
    tagnames = set();
    for src in sources:
      tagnames.update(src.getTagNames());
    self.tagnames = list(tagnames);
    self.tagnames.sort();

  def initGroupings (self):
    # init default and "selected" groupings
    # For the default style, make sure all style fields are initialied to proper values, so that some style setting is always guaranteed.
    # Do this by sarting with the Baseline style, and applying the specified default style to it as an update.
    if 'default' in self.plotstyles:
      defstyle = PlotStyles.BaselinePlotStyle.copy();
      defstyle.update(self.plotstyles['default']);
      defstyle.apply = 1000;  # apply at lowest priority
    else:
      defstyle = self.plotstyles['default'] = PlotStyles.BaselinePlotStyle;
    self.defgroup = Source.Grouping("all sources",func=lambda src:True,sources=self.sources,style=defstyle);
    self.curgroup = Source.Grouping("current source",func=lambda src:self.currentSource() is src,sources=self.sources,
                                                        style=self.plotstyles.setdefault('current',PlotStyles.HighlightPlotStyle));
    self.selgroup = Source.Grouping("selected sources",func=lambda src:getattr(src,'selected',False),sources=self.sources,
                                                        style=self.plotstyles.setdefault('selected',PlotStyles.SelectionPlotStyle));
    # and make ordered list of groupings
    self.groupings = [ self.defgroup,self.curgroup,self.selgroup ];
    # make groupings from available source types
    self._typegroups = {};
    typecodes = list(set([src.typecode for src in self.sources]));
    typecodes.sort();
    if len(typecodes) > 1:
      for code in typecodes:
          self._typegroups[code] = group = Source.Grouping("type: %s"%code,lambda src,code=code:src.typecode==code,sources=self.sources,
                                                                            style=self.plotstyles.setdefault('type:%s'%code,PlotStyles.DefaultPlotStyle));
          self.groupings.append(group);
    # make groupings from source tags
    self._taggroups = {};
    for tag in self.tagnames:
      self._taggroups[tag] = group = Source.Grouping("tag: %s"%tag,
                                                                    lambda src,tag=tag:getattr(src,tag,None) not in [None,False],
                                                                    sources=self.sources,
                                                                    style=self.plotstyles.setdefault('tag:%s'%tag,PlotStyles.DefaultPlotStyle));
      self.groupings.append(group);

  def _remakeGroupList (self):
    self.groupings = [ self.defgroup,self.curgroup,self.selgroup ];
    typenames = self._typegroups.keys();
    typenames.sort();
    self.groupings += [ self._typegroups[name] for name in typenames ];
    self.groupings += [ self._taggroups[name] for name in self.tagnames ];

  def getTagGrouping (self,tag):
    return self._taggroups[tag];

  def getTypeGrouping (self,typename):
    return self._typegroups[typename];

  def getSourcePlotStyle (self,src):
    """Returns PlotStyle object for given source, using the styles in the model grouping.
    Returns tuple of plotstyle,label, or None,None if no source is to be plotted.
    """;
    # get list of styles from  groupings to which this source belongs
    styles = [ group.style for group in self.groupings if group.func(src) ];
    # sort in order of priority (high apply to low apply)
    styles.sort(lambda a,b:cmp(b.apply,a.apply));
    # "show_plot" attribute: if at least one group is showing explicitly, show
    # else if at least one group is hiding explicitly, hide
    # else use default setting
    show = [ st.show_plot for st in styles ];
    if show and max(show) == PlotStyles.ShowAlways:
      show = True;
    elif show and min(show) == PlotStyles.ShowNot:
      show = False;
    else:
      show = bool(style0.show_plot);
    if not show:
      return None,None;
    # sort styles
    # Override attributes in style object with non-default attributes found in each matching grouping
    # Go in reverse, so 'current' overrides 'selected' overrides types overrides tags
    style = None;
    for st in styles:
      if st.apply:
        # make copy-on-write, so we don't overwrite the original style object
        if style is None:
          style = st.copy();
        else:
          style.update(st);
    return style,PlotStyles.makeSourceLabel(style.label,src);

  def addTag (self,tag):
    if tag in self.tagnames:
      return False;
    # tags beginning with "_" are internal, not added to tagname list
    if tag[0] == "_":
      return False;
    # add to list
    self.tagnames.append(tag);
    self.tagnames.sort();
    # add to groupings
    self._taggroups[tag] = Source.Grouping("tag: %s"%tag,
        lambda src,tag=tag:getattr(src,tag,None) not in [None,False],
        sources=self.sources,
        style=self.plotstyles.setdefault('tag:%s'%tag,PlotStyles.DefaultPlotStyle));
    # reform grouping list
    self._remakeGroupList();
    return True;

  def setFieldCenter (self,ra0,dec0):
    self.ra0,self.dec0 = ra0,dec0;

  def setPrimaryBeam (self,pbexp):
    self.pbexp = pbexp;

  def primaryBeam (self):
    return getattr(self,'pbexp',None);

  def setRefFreq (self,freq0):
    self.freq0 = freq0;

  def refFreq (self):
    return self.freq0;

  def hasFieldCenter (self):
    return self.ra0 is not None and self.dec0 is not None;

  def fieldCenter (self):
    """Returns center of field. If this is not explicitly specified in the model, uses the average position of all sources.""";
    if self.ra0 is None:
      self.ra0 = reduce(lambda x,y:x+y,[ src.pos.ra for src in self.sources ])/len(self.sources) if self.sources else 0;
    if self.dec0 is None:
      self.dec0 = reduce(lambda x,y:x+y,[ src.pos.dec for src in self.sources ])/len(self.sources)  if self.sources else 0;
    return self.ra0,self.dec0;

  def save (self,filename,format=None, verbose=True):
    """Convenience function, saves model to file. Format may be specified explicitly, or determined from filename.""";
    import Formats
    Formats.save(self,filename,format=format, verbose=verbose);

  _re_bynumber = re.compile("^([!-])?(\\d+)?:(\\d+)?$");

  def getSourcesNear (self,ra,dec,tolerance=DEG/60):
    return [ src for src in self.sources if angular_dist_pos_angle(src.pos.ra,src.pos.dec,ra,dec)[0]<tolerance ];

  def getSourceSubset (self,selection=None):
    """Gets list of sources matching the given selection string (if None, then all sources are returned.)""";
    if not selection or selection.lower() == "all":
      return self.sources;
    # sort by brightness
    srclist0 = sorted(self.sources,lambda a,b:cmp(b.brightness(),a.brightness()));
    all = set([src.name for src in srclist0]);
    srcs = set();
    for ispec,spec in enumerate(re.split("\s+|,",selection)):
      spec = spec.strip();
      if spec:
        # if first spec is a negation, then implictly select all sources first
        if not ispec and spec[0] in "!-":
          srcs = all;
        if spec.lower() == "all":
          srcs = all;
        elif self._re_bynumber.match(spec):
          negate,start,end = self._re_bynumber.match(spec).groups();
          sl = slice(int(start) if start else None,int(end) if end else None);
          if negate:
            srcs.difference_update([ src.name for src in srclist0[sl]]);
          else:
            srcs.update([ src.name for src in srclist0[sl]]);
        elif spec.startswith("-=") or spec.startswith("!="):
          srcs.difference_update([ src.name for src in srclist0 if getattr(src,spec[2:],None) ]);
        elif spec.startswith("="):
          srcs.update([ src.name for src in srclist0 if getattr(src,spec[1:],None) ]);
        elif spec.startswith("-")  or spec.startswith("!"):
          srcs.discard(spec[1:]);
        else:
          srcs.add(spec);
    # make list
    return [ src for src in srclist0 if src.name in srcs ];


SkyModel.registerClass();



#!/usr/bin/env python

from __future__ import print_function
from setuptools import setup, find_packages

__version__ = "1.4.2"

requirements = ['astro_kittens', 'numpy', 'scipy', 'astlib', 'pyfits']

scripts = [
    'Tigger/bin/tigger-convert',
    'Tigger/bin/tigger-make-brick',
    'Tigger/bin/tigger-restore',
    'Tigger/bin/tigger-tag',
]

package_data = {
}

extras_require = {
}


setup(
    name ="astro-tigger-lsm",
    version=__version__,
    packages=find_packages(),
    extras_require=extras_require,
    scripts=scripts,
    package_data=package_data,
    description="Python libraries and command-line tools for manipulating Tigger-format LSMs",
    author="Oleg Smirnov",
    author_email="osmirnov@gmail.com",
    url="https://github.com/ska-sa/tigger-lsm",
    install_requires=requirements,
)




# -*- coding: utf-8 -*-
#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys,re

import Kittens.utils

from Tigger.Models import ModelClasses
from Tigger.Models import SkyModel
from Tigger import Coordinates
import Tigger.Models.Formats
from Tigger.Models.Formats import dprint,dprintf,ASCII

"""Loads a PyBDSM-format .gaul file. Gaul files are essentially ASCII tables with a very specific naming convention."""

# Gaus_id Isl_id Source_id Wave_id RA E_RA DEC E_DEC Total_flux
# E_Total_flux Peak_flux E_Peak_flux Xposn E_Xposn Yposn E_Yposn Maj E_Maj Min
# E_Min PA E_PA DC_Maj E_DC_Maj DC_Min E_DC_Min DC_PA E_DC_PA Isl_Total_flux
# E_Isl_Total_flux Isl_rms Isl_mean Resid_Isl_rms Resid_Isl_mean S_Code

format_mapping = dict(
  Gaus_id="name",
  RA="ra_d",E_RA="ra_err_d",DEC="dec_d",E_DEC="dec_err_d",
  Total_flux="i",E_Total_flux="i_err",
  Total_Q="q",E_Total_Q="q_err",
  Total_U="u",E_Total_U="u_err",
  Total_V="v",E_Total_V="v_err",
  DC_Maj="emaj_d",DC_Min="emin_d",DC_PA="pa_d",
  E_DC_Maj="emaj_err_d",E_DC_Min="emin_err_d",E_DC_PA="pa_err_d",
  SpI="spi",Spec_Indx="spi",E_Spec_Indx="spi_err",
  S_Code=":str:_pybdsm_S_Code"
);


def load (filename, freq0=None,**kw):
  """Imports a gaul table
  The 'freq0' argument supplies a default reference frequency (if one is not contained in the file.)
  If 'center_on_brightest' is True, the mpodel field center will be set to the brightest source.
  'min_extent' is minimal source extent (in radians), above which a source will be treated as a Gaussian rather than a point component.
  """
  srclist = [];
  id = None
  dprint(1,"importing PyBDSM gaul/srl file",filename);
  format = {};
  extension = filename.split(".")[-1]
  if extension == "srl":
    format_mapping['Source_id'] = format_mapping.pop('Gaus_id')
    id = "Source_id"
  # look for format string and reference freq, and build up format dict
  for line in file(filename):
    m = re.match("# Reference frequency .*?([0-9.eE+-]+)\s*Hz",line);
    if m:
      freq0 = kw['freq0'] = freq0 or float(m.group(1));
      dprint(2,"found reference frequency %g"%freq0);
    elif re.match("#(\s*[\w:]+\s+)+",line) and line.find(id if id else "Gaus_id") > 0:
      dprint(2,"found format string",line);
      fields = dict([ (name,i) for i,name in enumerate(line[1:].split()) ]); 
      # map known fields to their ASCII equivalents, the rest copy as custom float attributes with
      # a "pybdsm_" prefix
      for i,name in enumerate(line[1:].split()):
        if name in format_mapping:
          dprint(2,"Field",format_mapping[name],name,"is column",i)
          format[format_mapping[name]] = i;
        else:
          format[":float:_pybdsm_%s"%name] = i;
    if format and freq0:
      break;
  if not format:
    raise ValueError,"this .gaul file does not appear to contain a format string"
  # call ASCII.load() function now that we have the format dict
  kw['format'] = format;
  return ASCII.load(filename,**kw)

Tigger.Models.Formats.registerFormat("Gaul",load,"PyBDSM .gaul/.srl file",(".gaul",".srl",));



# -*- coding: utf-8 -*-
# gaussfitter.py
# created by Adam Ginsburg (adam.ginsburg@colorado.edu or keflavich@gmail.com) 3/17/08)
#
#% $Id$ 
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from numpy import *
from scipy import optimize
from scipy import stats

def moments (data,circle,rotate,vheight):
    """Returns (height, amplitude, x, y, width_x, width_y, rotation angle)
    the gaussian parameters of a 2D distribution by calculating its
    moments.  Depending on the input parameters, will only output 
    a subset of the above"""
    total = data.sum()
    X, Y = indices(data.shape)
    x = (X*data).sum()/total
    y = (Y*data).sum()/total
    col = data[:, int(y)]
    width_x = sqrt(abs((arange(col.size)-y)**2*col).sum()/col.sum())
    row = data[int(x), :]
    width_y = sqrt(abs((arange(row.size)-x)**2*row).sum()/row.sum())
    width = ( width_x + width_y ) / 2.
    height = stats.mode(data.ravel())[0][0] if vheight else 0;
    amplitude = data.max()-height
    mylist = [amplitude,x,y]
    if vheight==1:
        mylist = [height] + mylist
    if circle==0:
        mylist = mylist + [width_x,width_y]
    else:
        mylist = mylist + [width]
    if rotate==1:
        mylist = mylist + [0.] #rotation "moment" is just zero...
    return tuple(mylist)

def twodgaussian(inpars, circle, rotate, vheight):
    """Returns a 2d gaussian function of the form:
        x' = cos(rota) * x - sin(rota) * y
        y' = sin(rota) * x + cos(rota) * y
        (rota should be in degrees)
        g = b + a exp ( - ( ((x-center_x)/width_x)**2 +
        ((y-center_y)/width_y)**2 ) / 2 )

        where x and y are the input parameters of the returned function,
        and all other parameters are specified by this function

        However, the above values are passed by list.  The list should be:
        inpars = (height,amplitude,center_x,center_y,width_x,width_y,rota)

        You can choose to ignore / neglect some of the above input parameters using the following options:
            circle=0 - default is an elliptical gaussian (different x, y widths), but can reduce
                the input by one parameter if it's a circular gaussian
            rotate=1 - default allows rotation of the gaussian ellipse.  Can remove last parameter
                by setting rotate=0
            vheight=1 - default allows a variable height-above-zero, i.e. an additive constant
                for the Gaussian function.  Can remove first parameter by setting this to 0
        """
    inpars_old = inpars
    inpars = list(inpars)
    if vheight == 1:
        height = inpars.pop(0)
        height = float(height)
    else:
        height = float(0)
    amplitude, center_x, center_y = inpars.pop(0),inpars.pop(0),inpars.pop(0)
    amplitude = float(amplitude)
    center_x = float(center_x)
    center_y = float(center_y)
    if circle == 1:
        width = inpars.pop(0)
        width_x = float(width)
        width_y = float(width)
    else:
        width_x, width_y = inpars.pop(0),inpars.pop(0)
        width_x = float(width_x)
        width_y = float(width_y)
    if rotate == 1:
        rota = inpars.pop(0)
        rota = pi/180. * float(rota)
        rcen_x = center_x * cos(rota) - center_y * sin(rota)
        rcen_y = center_x * sin(rota) + center_y * cos(rota)
    else:
        rcen_x = center_x
        rcen_y = center_y
    if len(inpars) > 0:
        raise ValueError("There are still input parameters:" + str(inpars) + \
                " and you've input: " + str(inpars_old) + " circle=%d, rotate=%d, vheight=%d" % (circle,rotate,vheight) )
            
    def rotgauss(x,y):
        if rotate==1:
            xp = x * cos(rota) - y * sin(rota)
            yp = x * sin(rota) + y * cos(rota)
        else:
            xp = x
            yp = y
        g = height+amplitude*exp(
            -(((rcen_x-xp)/width_x)**2+
            ((rcen_y-yp)/width_y)**2)/2.)
        return g
    return rotgauss

def gaussfit(data,err=None,params=[],autoderiv=1,return_all=0,circle=0,rotate=1,vheight=1):
    """
    Gaussian fitter with the ability to fit a variety of different forms of 2-dimensional gaussian.
    
    Input Parameters:
        data - 2-dimensional data array
        err=None - error array with same size as data array
        params=[] - initial input parameters for Gaussian function.
            (height, amplitude, x, y, width_x, width_y, rota)
            if not input, these will be determined from the moments of the system, 
            assuming no rotation
        autoderiv=1 - use the autoderiv provided in the lmder.f function (the alternative
            is to us an analytic derivative with lmdif.f: this method is less robust)
        return_all=0 - Default is to return only the Gaussian parameters.  See below for
            detail on output
        circle=0 - default is an elliptical gaussian (different x, y widths), but can reduce
            the input by one parameter if it's a circular gaussian
        rotate=1 - default allows rotation of the gaussian ellipse.  Can remove last parameter
            by setting rotate=0
        vheight=1 - default allows a variable height-above-zero, i.e. an additive constant
            for the Gaussian function.  Can remove first parameter by setting this to 0

    Output:
        Default output is a set of Gaussian parameters with the same shape as the input parameters
        Can also output the covariance matrix, 'infodict' that contains a lot more detail about
            the fit (see scipy.optimize.leastsq), and a message from leastsq telling what the exit
            status of the fitting routine was

        Warning: Does NOT necessarily output a rotation angle between 0 and 360 degrees.
    """
    if params == []:
        params = (moments(data,circle,rotate,vheight))
    if err == None:
        errorfunction = lambda p: ravel((twodgaussian(p,circle,rotate,vheight)(*indices(data.shape)) - data))
    else:
        errorfunction = lambda p: ravel((twodgaussian(p,circle,rotate,vheight)(*indices(data.shape)) - data)/err)
    if autoderiv == 0:
        # the analytic derivative, while not terribly difficult, is less efficient and useful.  I only bothered
        # putting it here because I was instructed to do so for a class project - please ask if you would like 
        # this feature implemented
        raise ValueError("I'm sorry, I haven't implemented this feature yet.")
    else:
        p, cov, infodict, errmsg, success = optimize.leastsq(errorfunction, params, full_output=1)
    if  return_all == 0:
        return p
    elif return_all == 1:
        return p,cov,infodict,errmsg



# -*- coding: utf-8 -*-
#
#% $Id$ 
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import ModelClasses
import math

# string used to indicate default value of an attribute
DefaultValue = "default";
# string used to indicate "none" value of an attribute
NoneValue = "none";

# definitive list of style attributes
StyleAttributes = [ "symbol","symbol_color","symbol_size","symbol_linewidth","label","label_color","label_size" ];

# dict of attribute labels (i.e. for menus, column headings and such)
StyleAttributeLabels = dict(symbol="symbol",symbol_color="color",symbol_size="size",symbol_linewidth="line width",
                                           label="label",label_color="color",label_size="size");
# dict of attribute types. Any attribute not in this dict is of type str.
StyleAttributeTypes = dict(symbol_size=int,symbol_linewidth=int,label_size=int);

# list of known colors
ColorList = [ "black","blue","lightblue","green","lightgreen","cyan","red","orange red","purple","magenta","yellow","white" ];
DefaultColor = "black";

# dict and method to pick a contrasting color (i.e. suitable as background for specified color)
ContrastColor = dict(white="#404040",yellow="#404040");
DefaultContrastColor = "#B0B0B0";

def getContrastColor (color):
  return ContrastColor.get(color,DefaultContrastColor);


# dict of possible user settings for each attribute
StyleAttributeOptions = dict(
  symbol = [ DefaultValue,NoneValue,"cross","plus","dot","circle","square","diamond" ],
  symbol_color =  [ DefaultValue ] + ColorList,
  label = [ DefaultValue,NoneValue,"%N","%N %BJy","%N %BJy r=%R'" ],
  label_color =  [ DefaultValue ] + ColorList,
  label_size = [ DefaultValue,6,8,10,12,14 ],
);

# constants for the show_list and show_plot attributes
ShowNot = 0;
ShowDefault = 1;
ShowAlways = 2;


DefaultPlotAttrs = dict(symbol=None,symbol_color=DefaultColor,symbol_size=5,symbol_linewidth=0,
                        label=None,label_color=DefaultColor,label_size=10,
                        show_list=ShowDefault,show_plot=ShowDefault,apply=0);

class PlotStyle (ModelClasses.ModelItem):
  optional_attrs = DefaultPlotAttrs;

  def copy (self):
    return PlotStyle(**dict([(attr,getattr(self,attr,default)) for attr,default in DefaultPlotAttrs.iteritems()]))

  def update (self,other):
    for attr in DefaultPlotAttrs.iterkeys():
      val = getattr(other,attr,None);
      if val is not None and val != DefaultValue:
        setattr(self,attr,val);

PlotStyle.registerClass();

# Default plot style. This must define everything! (I.e. no DefaultValue elements allowed.)
BaselinePlotStyle = PlotStyle(symbol="plus",symbol_color="yellow",symbol_size=2,symbol_linewidth=0,
                                            label=NoneValue,label_color="blue",label_size=6,
                                            show_list=ShowAlways,show_plot=ShowAlways,apply=1000);

SelectionPlotStyle = PlotStyle(symbol=DefaultValue,symbol_color="cyan",symbol_size=DefaultValue,symbol_linewidth=DefaultValue,
                                            label="%N",label_color="green",label_size=DefaultValue,
                                            show_list=ShowAlways,show_plot=ShowAlways,apply=-1);

HighlightPlotStyle = PlotStyle(symbol=DefaultValue,symbol_color="red",symbol_size=DefaultValue,symbol_linewidth=DefaultValue,
                                            label="%N %BJy",label_color="red",label_size=12,
                                            show_list=ShowAlways,show_plot=ShowAlways,apply=-2);

DefaultPlotStyle = PlotStyle(symbol=DefaultValue,symbol_color=DefaultValue,symbol_size=DefaultValue,symbol_linewidth=DefaultValue,
                                            label=DefaultValue,label_color=DefaultValue,label_size=DefaultValue,
                                            show_list=ShowDefault,show_plot=ShowDefault,apply=1000);

# cache of precompiled labels
_compiled_labels = {};

# label replacements
_label_keys = {   "%N": lambda src:src.name,
                            "%B": lambda src:"%.2g"%src.brightness(),
                            "%R": lambda src:(hasattr(src,'r') and "%.2g"%(src.r/math.pi*180*60)) or "",
                            "%T": lambda src:src.typecode,
                            "%I": lambda src:"%.2g"%getattr(src.flux,'I',0),
                            "%Q": lambda src:"%.2g"%getattr(src.flux,'Q',0),
                            "%U": lambda src:"%.2g"%getattr(src.flux,'U',0),
                            "%V": lambda src:"%.2g"%getattr(src.flux,'V',0),
};

def makeSourceLabel (label,src):
  if label == NoneValue or label is None:
    return "";
  global _label_keys;
  lbl = label;
  for key,func in _label_keys.iteritems():
    if lbl.find(key) >= 0:
      lbl = lbl.replace(key,func(src));
  return lbl;
